/*
** Copyright (c) 1995-1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
**
** File Name:   DDSLIBLT.C
**
** Description:	Internal DDRAW blt functions for SLI/AA mode
**	
*/

/****************************************************************************
*
* INTERNAL FUNCTIONS:
*
* sli_ColorFill       --- fill surface with a color, always use SRCCOPY ROP
* sli_PuntColorFill   --- fill surface with a color, always use SRCCOPY ROP
* sli_PuntColorFillZ  --- fill Z surface with a color, always use SRCCOPY ROP
* sli_DoBltNoSP       --- blt w/o source/pattern, may have dest. color key
* sli_DoBltS          --- screen to screen blt, also handle stretch/shrink
* sli_SystemToVideo   --- perform host to screen blt
*
****************************************************************************/

#include "precomp.h"

#if defined(SLI_AA)
#include <ddsli2d.h>

#pragma intrinsic (memcpy, memset)

/*
 * We use one scanline buffer to read the source because it gains performance,
 * and we do not need to special case overlapping in the x direction.
 *
 * We only use the second buffer if we need to read the destination before
 * writing back. Otherwise, we should not use the second buffer as it is
 * slower.
 */
// maximum x resolution: 2048, maximum color depth: 4 bytes per pixel

BYTE scanBuf1[2048* 4];
BYTE scanBuf2[2048* 4];



/***************************************************************************/
/*                             DEFINES                                     */
/***************************************************************************/

// pointers to functions returning voids

typedef void (__stdcall * PBLTFUNC) (BLT_PARAMS);
typedef void (__stdcall * PSBLTFUNC) (LPDDHAL_BLTDATA, BYTE *, BYTE *, DWORD, DWORD, DWORD, DWORD);
typedef void (* PFILLSCANFUNC) (BYTE *, BYTE *, DWORD);

typedef enum tag_colorKey { none, src, dst, both } CKEY_TYPE;

// not implemented features

#ifdef DEBUG
 #define NOT_SUPPORTED  _asm { int 3 }
#else
 #define NOT_SUPPORTED
#endif

// macro to convert SSTG_PIX_FMT_* to bytes per pixel

#define GET_BYTE_DEPTH(BytesPerPel, PelFormat) \
        switch(PelFormat) \
        { \
            case SSTG_PIXFMT_8BPP:  \
                BytesPerPel = 1;    \
                break;              \
            case SSTG_PIXFMT_16BPP: \
                BytesPerPel = 2;    \
                break;              \
            case SSTG_PIXFMT_32BPP: \
                BytesPerPel = 4;    \
                break;              \
            default: /* format not supported */ \
                BytesPerPel = 0;    \
                break;              \
        }

// macro to advance to the next scanline

#define NEXT_SCAN { srcPtr += srcPitch; dstPtr += dstPitch; }

/***************************************************************************/
/*                       FUNCTION PROTOTYPES                               */
/***************************************************************************/

// parameter list for blt functions

#define BLT_PARAMS \
    LPDDHAL_BLTDATA  pbd, \
    DWORD dstWidthBytes, DWORD dstHeight, DWORD srcPitch, DWORD dstPitch, \
    BYTE *srcPtr, BYTE *dstPtr, DWORD srcBytesPerPel, DWORD dstBytesPerPel

void __stdcall sBltSrcToScanBuf(LPDDHAL_BLTDATA, BYTE *, BYTE *, DWORD, DWORD, DWORD, DWORD);
void __stdcall xsBltSrcToScanBuf(LPDDHAL_BLTDATA,BYTE *, BYTE *, DWORD, DWORD, DWORD, DWORD);

void __stdcall bltDSa   (BLT_PARAMS);
void __stdcall bltDSan  (BLT_PARAMS);
void __stdcall bltDSna  (BLT_PARAMS);
void __stdcall bltDSno  (BLT_PARAMS);
void __stdcall bltDSo   (BLT_PARAMS);
void __stdcall bltDSon  (BLT_PARAMS);
void __stdcall bltDSx   (BLT_PARAMS);
void __stdcall bltDSxn  (BLT_PARAMS);
void __stdcall bltNop   (BLT_PARAMS);
void __stdcall bltS     (BLT_PARAMS);
void __stdcall bltSDna  (BLT_PARAMS);
void __stdcall bltSDno  (BLT_PARAMS);
void __stdcall bltSn    (BLT_PARAMS);

void __stdcall xBltDSa  (BLT_PARAMS);
void __stdcall xBltDSan (BLT_PARAMS);
void __stdcall xBltDSna (BLT_PARAMS);
void __stdcall xBltDSno (BLT_PARAMS);
void __stdcall xBltDSo  (BLT_PARAMS);
void __stdcall xBltDSon (BLT_PARAMS);
void __stdcall xBltDSx  (BLT_PARAMS);
void __stdcall xBltDSxn (BLT_PARAMS);
void __stdcall xBltS    (BLT_PARAMS);
void __stdcall xBltSDna (BLT_PARAMS);
void __stdcall xBltSDno (BLT_PARAMS);
void __stdcall xBltSn   (BLT_PARAMS);

void __stdcall sBltS    (BLT_PARAMS);

/***************************************************************************/
/*                         FUNCTION TABLES                                 */
/***************************************************************************/

// VRAM/system memory to VRAM opaque blt, no stretch/shrink

PBLTFUNC BltFuncTab[16] =
{
    (PBLTFUNC) &bltNop,     // BLACKNESS is supported by sli_DoBltNoSP()
    (PBLTFUNC) &bltDSon,    // NOTSRCERASE
    (PBLTFUNC) &bltDSna,
    (PBLTFUNC) &bltSn,      // NOTSRCCOPY
    (PBLTFUNC) &bltSDna,    // SRCERASE
    (PBLTFUNC) &bltNop,     // DSTINVERT is supported by sli_DoBltNoSP()
    (PBLTFUNC) &bltDSx,     // SRCINVERT
    (PBLTFUNC) &bltDSan,
    (PBLTFUNC) &bltDSa,     // SRCAND
    (PBLTFUNC) &bltDSxn,
    (PBLTFUNC) &bltNop,     // rop D is supported by sli_DoBltNoSP()
    (PBLTFUNC) &bltDSno,    // MERGEPAINT
    (PBLTFUNC) &bltS,       // SRCCOPY
    (PBLTFUNC) &bltSDno,
    (PBLTFUNC) &bltDSo,     // SRCPAINT
    (PBLTFUNC) &bltNop,     // WHITENESS is supported by sli_DoBltNoSP()
};

// VRAM/system memory to VRAM transparent blt, no stretch/shrink

PBLTFUNC xBltFuncTab[16] =
{
    (PBLTFUNC) &bltNop,     // BLACKNESS is supported by sli_DoBltNoSP()
    (PBLTFUNC) &xBltDSon,   // NOTSRCERASE
    (PBLTFUNC) &xBltDSna,
    (PBLTFUNC) &xBltSn,     // NOTSRCCOPY
    (PBLTFUNC) &xBltSDna,   // SRCERASE
    (PBLTFUNC) &bltNop,     // DSTINVERT is supported by sli_DoBltNoSP()
    (PBLTFUNC) &xBltDSx,    // SRCINVERT
    (PBLTFUNC) &xBltDSan,
    (PBLTFUNC) &xBltDSa,    // SRCAND
    (PBLTFUNC) &xBltDSxn,
    (PBLTFUNC) &bltNop,     // rop D is supported by sli_DoBltNoSP()
    (PBLTFUNC) &xBltDSno,   // MERGEPAINT
    (PBLTFUNC) &xBltS,      // SRCCOPY
    (PBLTFUNC) &xBltSDno,
    (PBLTFUNC) &xBltDSo,    // SRCPAINT
    (PBLTFUNC) &bltNop,     // WHITENESS is supported by sli_DoBltNoSP()
};

// VRAM/system memory to VRAM opaque blt, stretch/shrink

PBLTFUNC sBltFuncTab[16] =
{
    (PBLTFUNC) &bltNop,     // BLACKNESS is supported by sli_DoBltNoSP()
    (PBLTFUNC) &xBltDSon,   // NOTSRCERASE
    (PBLTFUNC) &xBltDSna,
    (PBLTFUNC) &xBltSn,     // NOTSRCCOPY
    (PBLTFUNC) &xBltSDna,   // SRCERASE
    (PBLTFUNC) &bltNop,     // DSTINVERT is supported by sli_DoBltNoSP()
    (PBLTFUNC) &xBltDSx,    // SRCINVERT
    (PBLTFUNC) &xBltDSan,
    (PBLTFUNC) &xBltDSa,    // SRCAND
    (PBLTFUNC) &xBltDSxn,
    (PBLTFUNC) &bltNop,     // rop D is supported by sli_DoBltNoSP()
    (PBLTFUNC) &xBltDSno,   // MERGEPAINT
    (PBLTFUNC) &sBltS,      // SRCCOPY
    (PBLTFUNC) &xBltSDno,
    (PBLTFUNC) &xBltDSo,    // SRCPAINT
    (PBLTFUNC) &bltNop,     // WHITENESS is supported by sli_DoBltNoSP()
};


/***************************************************************************/
/*                       INTERNAL FUNCTIONS                                */
/***************************************************************************/

void __stdcall xBltDSa(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSan(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSna(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSno(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSo(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSon(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSx(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltDSxn(BLT_PARAMS)
{
    NOT_SUPPORTED
}

void __stdcall xBltS(BLT_PARAMS)
{
    CKEY_TYPE colorKey;
    DWORD srcColorkeyMin, srcColorkeyMax;
    DWORD dstColorkeyMin, dstColorkeyMax;
    DWORD dwData, i;
    WORD  wData;
    BYTE *pSrcData, *pDstData, bData, *pData;

    colorKey = none;

    if (pbd->dwFlags & DDBLT_KEYSRCOVERRIDE)
    {
        colorKey = src;
        srcColorkeyMin = pbd->bltFX.ddckSrcColorkey.dwColorSpaceLowValue;
        srcColorkeyMax = pbd->bltFX.ddckSrcColorkey.dwColorSpaceHighValue;
    }

    if (pbd->dwFlags & DDBLT_KEYDESTOVERRIDE)
    {
        colorKey = (src==colorKey)? both : dst;
        dstColorkeyMin = pbd->bltFX.ddckDestColorkey.dwColorSpaceLowValue;
        dstColorkeyMax = pbd->bltFX.ddckDestColorkey.dwColorSpaceHighValue;
    }

    if (src == colorKey)    // do not write if src pixel matches color key
    {
        switch(srcBytesPerPel)
        {
        case 1: // 8bpp
            while (dstHeight--)
            {
                memcpy(scanBuf1, srcPtr, dstWidthBytes);
                pDstData = dstPtr;
                for (i=0; i < dstWidthBytes; i++)
                {
                    bData = scanBuf1[i];
                    if ((bData < srcColorkeyMin) || (bData > srcColorkeyMax))
                    {
                        *((BYTE *)pDstData) = bData;
                    }
                    ((BYTE *)pDstData)++;
                }
                NEXT_SCAN
            }
            break;

        case 2: // 16bpp
            while (dstHeight--)
            {
                memcpy(scanBuf1, srcPtr, dstWidthBytes);
                pDstData = dstPtr;
                pData = scanBuf1;
                for (i=0; i < (dstWidthBytes >> 1); i++)
                {
                    wData = *((WORD *)pData)++;
                    if ((wData < srcColorkeyMin) || (wData > srcColorkeyMax))
                    {
                        *((WORD *)pDstData) = wData;
                    }
                    ((WORD *)pDstData)++;
                }
                NEXT_SCAN
            }
            break;

        case 4: // 32bpp
            while (dstHeight--)
            {
                memcpy(scanBuf1, srcPtr, dstWidthBytes);
                pDstData = dstPtr;
                pData = scanBuf1;
                for (i=0; i < (dstWidthBytes >> 2); i++)
                {
                    dwData = *((DWORD *)pData)++ & 0xffffff;
                    if ((dwData < srcColorkeyMin) || (dwData > srcColorkeyMax))
                    {
                        *((DWORD *)pDstData) = dwData;
                    }
                    ((DWORD *)pDstData)++;
                }
                NEXT_SCAN
            }
            break;
        } // end: switch(bytesPerPel)
    } // endif: use src color key
    else
    {
        if (dst == colorKey ) // update dst. pixel if matches color key
        {
            switch(dstBytesPerPel)
            {
            case 1: // 8bpp
                while (dstHeight--)
                {
                    memcpy(scanBuf1, srcPtr, dstWidthBytes);
                    memcpy(scanBuf2, dstPtr, dstWidthBytes);
                    pSrcData = scanBuf1;
                    pData = scanBuf2;
                    pDstData = dstPtr;
                    for (i=0; i < dstWidthBytes; i++)
                    {
                        bData = *((BYTE *)pData)++;
                        if ((bData >= dstColorkeyMin) &&
                            (bData <= dstColorkeyMax)
                           )
                        {
                            *((BYTE *)pDstData) = *((BYTE *)pSrcData);
                        }
                        ((BYTE *)pDstData)++;
                        ((BYTE *)pSrcData)++;
                    }
                    NEXT_SCAN
                }
                break;

            case 2: // 16bpp
                while (dstHeight--)
                {
                    memcpy(scanBuf1, srcPtr, dstWidthBytes);
                    memcpy(scanBuf2, dstPtr, dstWidthBytes);
                    pSrcData = scanBuf1;
                    pData = scanBuf2;
                    pDstData = dstPtr;
                    for (i=0; i < (dstWidthBytes >> 1); i++)
                    {
                        wData = *((WORD *)pData)++;
                        if ((wData >= dstColorkeyMin) &&
                            (wData <= dstColorkeyMax)
                           )
                        {
                            *((WORD *)pDstData) = *((WORD *)pSrcData);
                        }
                        ((WORD *)pDstData)++;
                        ((WORD *)pSrcData)++;
                    }
                    NEXT_SCAN
                }
                break;

            case 4: // 32bpp
                while (dstHeight--)
                {
                    memcpy(scanBuf1, srcPtr, dstWidthBytes);
                    memcpy(scanBuf2, dstPtr, dstWidthBytes);
                    pSrcData = scanBuf1;
                    pData = scanBuf2;
                    pDstData = dstPtr;
                    for (i=0; i < (dstWidthBytes >> 2); i++)
                    {
                        dwData = *((DWORD *)pData)++ & 0xffffff;
                        if ((dwData >= dstColorkeyMin) &&
                            (dwData <= dstColorkeyMax)
                           )
                        {
                            *((DWORD *)pDstData) = *((DWORD *)pSrcData);
                        }
                        ((DWORD *)pDstData)++;
                        ((DWORD *)pSrcData)++;
                    }
                    NEXT_SCAN
                }
                break;
            } // end: switch(bytesPerPel)
        } // endif: use dst color key
        else
        {   // use both src and dst color key

            switch(dstBytesPerPel)
            {
            case 1:
                while (dstHeight--)
                {
                    memcpy(scanBuf1, srcPtr, dstWidthBytes);
                    memcpy(scanBuf2, dstPtr, dstWidthBytes);
                    pSrcData = scanBuf1;
                    pData = scanBuf2;
                    pDstData = dstPtr;
                    for (i=0; i < dstWidthBytes; i++)
                    {
                        bData = *((BYTE *)pSrcData);
                        if ((bData < srcColorkeyMin) || (bData > srcColorkeyMax))
                        {
                            bData = *((BYTE *)pData)++;
                            if ((bData >= dstColorkeyMin) &&
                                (bData <= dstColorkeyMax)
                               )
                            {
                                *((BYTE *)pDstData) = *((BYTE *)pSrcData);
                            }
                         }
                         ((BYTE *)pSrcData)++;
                         ((BYTE *)pDstData)++;
                    }
                    NEXT_SCAN
                }
                break;

            case 2:
                while (dstHeight--)
                {
                    memcpy(scanBuf1, srcPtr, dstWidthBytes);
                    memcpy(scanBuf2, dstPtr, dstWidthBytes);
                    pSrcData = scanBuf1;
                    pData = scanBuf2;
                    pDstData = dstPtr;
                    for (i=0; i < (dstWidthBytes >> 1); i++)
                    {
                        wData = *((WORD *)pSrcData);
                        if ((wData < srcColorkeyMin) || (wData > srcColorkeyMax))
                        {
                            wData = *((WORD *)pData)++;
                            if ((wData >= dstColorkeyMin) &&
                                (wData <= dstColorkeyMax)
                               )
                            {
                                *((WORD *)pDstData) = *((WORD *)pSrcData);
                            }
                         }
                         ((WORD *)pSrcData)++;
                         ((WORD *)pDstData)++;
                    }
                    NEXT_SCAN
                }
                break;

            case 4:
                while (dstHeight--)
                {
                    memcpy(scanBuf1, srcPtr, dstWidthBytes);
                    memcpy(scanBuf2, dstPtr, dstWidthBytes);
                    pSrcData = scanBuf1;
                    pData = scanBuf2;
                    pDstData = dstPtr;
                    for (i=0; i < (dstWidthBytes >> 1); i++)
                    {
                        dwData = *((DWORD *)pSrcData) & 0xffffff;
                        if ((dwData < srcColorkeyMin) || (dwData > srcColorkeyMax))
                        {
                            dwData = *((DWORD *)pData)++ & 0xffffff;
                            if ((dwData >= dstColorkeyMin) &&
                                (dwData <= dstColorkeyMax)
                               )
                            {
                                *((DWORD *)pDstData) = *((DWORD *)pSrcData);
                            }
                         }
                         ((DWORD *)pSrcData)++;
                         ((DWORD *)pDstData)++;
                    }
                    NEXT_SCAN
                }
                break;
            } // end: switch(bytesPerPel)
        } // endif: use src and dst color key
    } // endif: use src color key
}

void __stdcall xBltSDna(BLT_PARAMS)
{
}

void __stdcall xBltSDno(BLT_PARAMS)
{
}

void __stdcall xBltSn(BLT_PARAMS)
{
}

/*--------------------------------------------------------------------*/
void __stdcall bltDSa(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ &= *((DWORD *)pSrcData)++;
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ &= *((BYTE *)pSrcData)++;
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSan(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pSrcData)++ & *((DWORD *)pDstData) );
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pSrcData)++ & *((BYTE *)pDstData) );
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSna(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pSrcData)++ ) & *((DWORD *)pDstData);
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pSrcData)++ ) & *((BYTE *)pDstData);
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSno(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pSrcData)++ ) | *((DWORD *)pDstData);
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pSrcData)++ ) | *((BYTE *)pDstData);
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSo(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ |= *((DWORD *)pSrcData)++;
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ |= *((BYTE *)pSrcData)++;
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSon(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pSrcData)++ | *((DWORD *)pDstData) );
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pSrcData)++ | *((BYTE *)pDstData) );
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSx(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ ^= *((DWORD *)pSrcData)++;
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ ^= *((BYTE *)pSrcData)++;
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltDSxn(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pSrcData)++ ^ *((DWORD *)pDstData) );
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pSrcData)++ ^ *((BYTE *)pDstData) );
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltNop(BLT_PARAMS)
{
    /* nothing to do, hurry home */
}

void __stdcall bltS(BLT_PARAMS)
{
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(dstPtr, scanBuf1, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltSDna(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pDstData) ) & *((DWORD *)pSrcData)++;
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pDstData) ) & *((BYTE *)pSrcData)++;
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltSDno(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        memcpy(scanBuf2, dstPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = scanBuf2;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pDstData) ) | *((DWORD *)pSrcData)++;
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pDstData) ) | *((BYTE *)pSrcData)++;
        }
        memcpy(dstPtr, scanBuf2, dstWidthBytes);
        NEXT_SCAN
    }
}

void __stdcall bltSn(BLT_PARAMS)
{
    DWORD rightEdge, i;
    BYTE *pSrcData, *pDstData;

    rightEdge = dstWidthBytes & 3;
    while (dstHeight--)
    {
        memcpy(scanBuf1, srcPtr, dstWidthBytes);
        pSrcData = scanBuf1;
        pDstData = dstPtr;
        for (i=0; i < (dstWidthBytes >> 2); i++)
        {
            *((DWORD *)pDstData)++ =
                ~( *((DWORD *)pSrcData)++ );
        }
        for (i=0; i < rightEdge; i++)
        {
            *((BYTE *)pDstData)++ =
                ~( *((BYTE *)pSrcData)++ );
        }
        NEXT_SCAN
    }
}
/*--------------------------------------------------------------------*/

/*----------------------------------------------------------------------
Function name:  initBuffer

Description:    initialize the scanline buffer
----------------------------------------------------------------------*/
void initBuffer(BYTE * pBuf, DWORD bpp, DWORD color, DWORD dstWidthBytes)
{
  DWORD i;
  BYTE * pData;

  switch(bpp)
  {
    case 1: // 8bpp
      memset(pBuf, color, dstWidthBytes);
      break;
    case 2: // 16bpp
      pData = pBuf;
      for (i=0; i < (dstWidthBytes >> 1); i++)
      {
        *((WORD *)pData)++ = (WORD)color;
      }
      break;
    case 4: // 32bpp
      pData = pBuf;
      for (i=0; i < (dstWidthBytes >> 2); i++)
      {
        *((DWORD *)pData)++ = (DWORD)color;
      }
      break;
   }
}

/*----------------------------------------------------------------------
Function name:  fillScanAND

Description:    primitive to perform a bitwise AND operation on data
----------------------------------------------------------------------*/
void fillScanAND(BYTE *dstPtr, BYTE *srcPtr, DWORD dstWidthBytes)
{
  DWORD rightEdge, i;
  BYTE *pDstData;

  // read frame buffer into scanline buffer
  memcpy(scanBuf2, dstPtr, dstWidthBytes);

  // bitwise AND source data with scanline buffer
  rightEdge = dstWidthBytes & 3;
  pDstData = scanBuf2;
  for (i=0; i < (dstWidthBytes >> 2); i++)
  {
    *((DWORD *)pDstData)++ &= *((DWORD *)srcPtr)++;
  }
  for (i=0; i < rightEdge; i++)
  {
    *((BYTE *)pDstData)++ &= *((BYTE *)srcPtr)++;
  }

  // write scanline buffer back to frame buffer
  memcpy(dstPtr, scanBuf2, dstWidthBytes);
}

/*----------------------------------------------------------------------
Function name:  fillScanOR

Description:    primitive to perform a bitwise OR operation on data
----------------------------------------------------------------------*/
void fillScanOR(BYTE *dstPtr, BYTE *srcPtr, DWORD dstWidthBytes)
{
  DWORD rightEdge, i;
  BYTE *pDstData;

  // read frame buffer into scanline buffer
  memcpy(scanBuf2, dstPtr, dstWidthBytes);

  // bitwise OR source data in with scanline buffer
  rightEdge = dstWidthBytes & 3;
  pDstData = scanBuf2;
  for (i=0; i < (dstWidthBytes >> 2); i++)
  {
    *((DWORD *)pDstData)++ |= *((DWORD *)srcPtr)++;
  }
  for (i=0; i < rightEdge; i++)
  {
    *((BYTE *)pDstData)++ |= *((BYTE *)srcPtr)++;
  }

  // write scanline buffer back to frame buffer
  memcpy(dstPtr, scanBuf2, dstWidthBytes);
}

/*----------------------------------------------------------------------
Function name:  fillScanSRC

Description:    primitive to copy source data to destination
----------------------------------------------------------------------*/
void fillScanSRC(BYTE *dstPtr, BYTE *srcPtr, DWORD dstWidthBytes)
{
  memcpy(dstPtr, srcPtr, dstWidthBytes);
}

/*----------------------------------------------------------------------
Function name:  sli_ColorFill

Description:    Fill a rectangular region of a surface.

Return:         DWORD DDRAW result

                DDHAL_DRIVER_HANDLED
----------------------------------------------------------------------*/
DWORD __stdcall
sli_ColorFill(NT9XDEVICEDATA *ppdev, LPDDHAL_BLTDATA pbd, DWORD BytesPerPel)
{
    DWORD dstWidth, dstHeight, dstPitch, dstWidthBytes;
    BYTE  *dstPtr;
    LPDDRAWI_DDRAWSURFACE_LCL lpDstSurf;

    FXBUSYWAIT(ppdev);

    MODIFY_SLI_READ(ppdev, ENABLE_SLI_READ);

    dstWidth = pbd->rDest.right - pbd->rDest.left;
    dstHeight = pbd->rDest.bottom - pbd->rDest.top;
    lpDstSurf = pbd->lpDDDestSurface;
    dstWidthBytes = dstWidth * BytesPerPel;

    dstPitch = ( IS_TILED(GET_HW_ADDR(lpDstSurf)) ) ?
                _DS(ddTilePitch) : lpDstSurf->lpGbl->lPitch;

    dstPtr = (BYTE *)
        ( lpDstSurf->lpGbl->fpVidMem +
          (pbd->rDest.top * dstPitch) + (pbd->rDest.left * BytesPerPel)
        );
#ifdef WINNT
    (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif

    initBuffer(scanBuf1, BytesPerPel, pbd->bltFX.dwFillColor, dstWidthBytes);

    while (dstHeight--)
    {
        memcpy(dstPtr, scanBuf1, dstWidthBytes);
        dstPtr += dstPitch; // next scan
    }

  MODIFY_SLI_READ(ppdev, DISABLE_SLI_READ);

  pbd->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;

} // sli_ColorFill


/*----------------------------------------------------------------------
Function name:  sli_ColorFillZ

Description:    Fill the Z buffer using direct frame buffer write.  This
                routine assumes that we offset the start of the z buffer by
                one tile, and there is an extra row of tiles at the bottom.

Return:         DWORD DDRAW result

                DDHAL_DRIVER_HANDLED
----------------------------------------------------------------------*/
DWORD __stdcall
sli_ColorFillZ(NT9XDEVICEDATA *ppdev, LPDDHAL_BLTDATA pbd, DWORD BytesPerPel)
{
    DWORD dstWidth, dstHeight, dstWidthBytes, fillWidth;
    DWORD xStart, xEnd, wrapWidth, hwPitch, linearPitch, xAdvance;
    BYTE  *dstPtr, *wrapPtr;
    LPDDRAWI_DDRAWSURFACE_LCL lpDstSurf;

    FXBUSYWAIT(ppdev);

    MODIFY_SLI_READ(ppdev, ENABLE_SLI_READ);

    dstWidth = pbd->rDest.right - pbd->rDest.left;
    dstHeight = pbd->rDest.bottom - pbd->rDest.top;
    lpDstSurf = pbd->lpDDDestSurface;
    dstWidthBytes = dstWidth * BytesPerPel;

    hwPitch = _DS(ddTileStride) << SST_TILE_WIDTH_BITS;
    linearPitch = lpDstSurf->lpGbl->lPitch;

    dstPtr = (BYTE *)
        ( lpDstSurf->lpGbl->fpVidMem +
          (pbd->rDest.top * linearPitch) + (pbd->rDest.left * BytesPerPel)
        );
#ifdef WINNT
    (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif

    initBuffer(scanBuf1, BytesPerPel, pbd->bltFX.dwFillColor, dstWidthBytes);

    if ( IS_TILED(GET_HW_ADDR(lpDstSurf)) )
    {
      xStart = pbd->rDest.left * BytesPerPel;
      xEnd = pbd->rDest.right * BytesPerPel;
      wrapWidth = SST_TILE_WIDTH - 
                  (hwPitch - (lpDstSurf->lpGbl->wWidth * BytesPerPel));

      if ( (long)xEnd < (long)(hwPitch - wrapWidth) )
      {
        while (dstHeight--)
        {
          memcpy(dstPtr, scanBuf1, dstWidthBytes);
          dstPtr += linearPitch; // next scan
        }      
      }
      else
      {
        // Special case for pixels that belong to the same scanline but 
        // lie in a tile that is on the next row.

        fillWidth = (dstWidthBytes > wrapWidth)?
                    (hwPitch - wrapWidth - xStart) : 0;
        wrapWidth = dstWidthBytes - fillWidth;

        wrapPtr = (BYTE *) (lpDstSurf->lpGbl->fpVidMem + 
                            (pbd->rDest.top * linearPitch));

        if (_DD(ddSLIModeEnabled))
        {
          xAdvance = ((linearPitch << SST_TILE_HEIGHT_BITS) * _DS(dwNumUnits))
                      - SST_TILE_WIDTH;
        }
        else
        {
          xAdvance = (linearPitch << SST_TILE_HEIGHT_BITS) - SST_TILE_WIDTH;
        }

        while (dstHeight--)
        {
          memcpy(dstPtr, scanBuf1, fillWidth);
          memcpy((wrapPtr + xAdvance), scanBuf1, wrapWidth);
          dstPtr += linearPitch; // next scan
          wrapPtr += linearPitch; // next scan
        }      
      }
    }
    else // linear Z buffer fill
    {
      while (dstHeight--)
      {
        memcpy(dstPtr, scanBuf1, dstWidthBytes);
        dstPtr += linearPitch; // next scan
      }
    }

    MODIFY_SLI_READ(ppdev, DISABLE_SLI_READ);

    pbd->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;

} // sli_ColorFillZ


/*----------------------------------------------------------------------
Function name:  sli_PuntColorFillZ

Description:    Fill the Z buffer using direct frame buffer write.  This
                routine assumes that we offset the start of the z buffer by
                one tile, and there is an extra row of tiles at the bottom.

Return:         DWORD DDRAW result

                DDHAL_DRIVER_HANDLED
----------------------------------------------------------------------*/
DWORD __stdcall
sli_PuntColorFillZ(NT9XDEVICEDATA *ppdev, PSOLIDCOLORPARAMS pParams)
{
    DWORD dstWidth, dstHeight, dstWidthBytes, fillWidth;
    DWORD xStart, xEnd, wrapWidth, hwPitch, linearPitch, xAdvance;
    BYTE  *dstPtr, *wrapPtr;
    PFILLSCANFUNC fillScan;

    FXBUSYWAIT(ppdev);

    MODIFY_SLI_READ(ppdev, ENABLE_SLI_READ);

    dstWidth = pParams->dstRight - pParams->dstLeft;
    dstHeight = pParams->dstBottom - pParams->dstTop;
    dstWidthBytes = dstWidth * pParams->BytesPerPel;

    hwPitch = _DS(ddTileStride) << SST_TILE_WIDTH_BITS;
    linearPitch = pParams->linearPitch;

    dstPtr = (BYTE *)
        ( pParams->fpVidMem + (pParams->dstTop * linearPitch) + 
          (pParams->dstLeft * pParams->BytesPerPel)
        );
#ifdef WINNT
    (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif

    initBuffer(scanBuf1, pParams->BytesPerPel, pParams->fillData, 
      dstWidthBytes);

    switch (pParams->bltRop)
    {
      case ((SSTG_ROP_AND << 16) | (SSTG_ROP_AND << 8) | (SSTG_ROP_AND)):
        fillScan = (PFILLSCANFUNC) &fillScanAND;
        break;
      case ((SSTG_ROP_OR << 16) | (SSTG_ROP_OR << 8) | (SSTG_ROP_OR)):
        fillScan = (PFILLSCANFUNC) &fillScanOR;
        break;
      case ((SSTG_ROP_SRC << 16) | (SSTG_ROP_SRC << 8) | (SSTG_ROP_SRC)):
      default:
        fillScan = (PFILLSCANFUNC) &fillScanSRC;
        break;
    }

    if (pParams->isTiled)
    {
      xStart = pParams->dstLeft * pParams->BytesPerPel;
      xEnd = pParams->dstRight * pParams->BytesPerPel;
      wrapWidth = SST_TILE_WIDTH - 
#ifdef WINNT
                  (hwPitch - (ppdev->cyScreen * pParams->BytesPerPel));
#else
                  (hwPitch - (_DS(hres) * pParams->BytesPerPel));
#endif

      if ( (long)xEnd < (long)(hwPitch - wrapWidth) )
      {
        while (dstHeight--)
        {
          (* fillScan)(dstPtr, scanBuf1, dstWidthBytes);
          dstPtr += linearPitch; // next scan
        }      
      }
      else
      {
        // Special case for pixels that belong to the same scanline but 
        // lie in a tile that is on the next row.

        fillWidth = (dstWidthBytes > wrapWidth)?
                    (hwPitch - wrapWidth - xStart) : 0;
        wrapWidth = dstWidthBytes - fillWidth;

        wrapPtr = (BYTE *) (pParams->fpVidMem + (pParams->dstTop * linearPitch));

        if (_DD(ddSLIModeEnabled))
        {
          xAdvance = ((linearPitch << SST_TILE_HEIGHT_BITS) * _DS(dwNumUnits)) - SST_TILE_WIDTH;
        }
        else
        {
          xAdvance = (linearPitch << SST_TILE_HEIGHT_BITS) - SST_TILE_WIDTH;
        }

        while (dstHeight--)
        {
          (* fillScan)(dstPtr, scanBuf1, fillWidth);
          (* fillScan)((wrapPtr + xAdvance), scanBuf1, wrapWidth);
          dstPtr += linearPitch; // next scan
          wrapPtr += linearPitch; // next scan
        }      
      }
    }
    else // linear Z buffer fill
    {
      while (dstHeight--)
      {
        (* fillScan)(dstPtr, scanBuf1, dstWidthBytes);
        dstPtr += linearPitch; // next scan
      }
    }

    MODIFY_SLI_READ(ppdev, DISABLE_SLI_READ);

    return DDHAL_DRIVER_HANDLED;

} // sli_PuntColorFillZ

/*----------------------------------------------------------------------
Function name:  sli_PuntColorFill

Description:    Fill a rectangular region of a surface.

Return:         DWORD DDRAW result

                DDHAL_DRIVER_HANDLED
----------------------------------------------------------------------*/
DWORD __stdcall 
sli_PuntColorFill(NT9XDEVICEDATA *ppdev, PSOLIDCOLORPARAMS pParams)
{
    DWORD dstWidth, dstHeight, dstPitch, dstWidthBytes;
    BYTE  *dstPtr;

    FXBUSYWAIT(ppdev);

    MODIFY_SLI_READ(ppdev, ENABLE_SLI_READ);

    dstWidth = pParams->dstRight - pParams->dstLeft;
    dstHeight = pParams->dstBottom - pParams->dstTop;
    dstWidthBytes = dstWidth * pParams->BytesPerPel;

    dstPitch = pParams->Pitch;

    dstPtr = (BYTE *)
        ( pParams->fpVidMem +
          (pParams->dstTop * dstPitch) + 
          (pParams->dstLeft * pParams->BytesPerPel)
        );
#ifdef WINNT
    (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif

    initBuffer(scanBuf1, pParams->BytesPerPel, pParams->fillData, 
      dstWidthBytes);

    while (dstHeight--)
    {
        memcpy(dstPtr, scanBuf1, dstWidthBytes);
        dstPtr += dstPitch; // next scan
    }

  MODIFY_SLI_READ(ppdev, DISABLE_SLI_READ);

  return DDHAL_DRIVER_HANDLED;

} // sli_PuntColorFill

/*----------------------------------------------------------------------
Function name:  sli_DoBltNoSP

Description:    Handle blts that don't use a source or pattern.

Return:         DWORD DDRAW result

                DDHAL_DRIVER_HANDLED
----------------------------------------------------------------------*/
DWORD __stdcall
sli_DoBltNoSP(NT9XDEVICEDATA   *ppdev,
              LPDDHAL_BLTDATA   pbd,
              DWORD             rop3,
              DWORD             BytesPerPel)
{
  DWORD dstWidth, dstHeight, dstPitch, dstWidthBytes;
  DWORD rightEdge, i;
  BYTE  *dstPtr, *pData;
  LPDDRAWI_DDRAWSURFACE_LCL lpDstSurf;

  FXBUSYWAIT(ppdev);

  MODIFY_SLI_READ(ppdev, ENABLE_SLI_READ);

  if (pbd->dwFlags & DDBLT_KEYDESTOVERRIDE)
  {
    // Software path cannot handle destination colorkey.

    pbd->ddRVal = DDERR_UNSUPPORTED;
    return DDHAL_DRIVER_HANDLED;
  }
  else
  {
    switch(rop3)
    {
    case ( HIWORD(BLACKNESS) ):
        pbd->bltFX.dwFillColor = rop3;
        sli_ColorFill(ppdev, pbd, BytesPerPel);
        break;

    case ( HIWORD(WHITENESS) ):
        pbd->bltFX.dwFillColor = 0xffffffff;
        sli_ColorFill(ppdev, pbd, BytesPerPel);
        break;

    case ( HIWORD(DSTINVERT) ):
        dstWidth = pbd->rDest.right - pbd->rDest.left;
        dstHeight = pbd->rDest.bottom - pbd->rDest.top;
        lpDstSurf = pbd->lpDDDestSurface;
        dstWidthBytes = dstWidth * BytesPerPel;

        dstPitch = ( IS_TILED(GET_HW_ADDR(lpDstSurf)) ) ?
            _DS(ddTilePitch) : lpDstSurf->lpGbl->lPitch;

        dstPtr = (BYTE *)
            ( lpDstSurf->lpGbl->fpVidMem +
              (pbd->rDest.top * dstPitch) + (pbd->rDest.left * BytesPerPel)
            );
#ifdef WINNT
        (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif

        rightEdge = dstWidthBytes & 3;

        while (dstHeight--)
        {
            memcpy(scanBuf1, dstPtr, dstWidthBytes);
            pData = scanBuf1;
            for (i=0; i < (dstWidthBytes >> 2); i++)
            {
                *((DWORD *)pData)++ = ~(*((DWORD *)pData));
            }
            for (i=0; i < rightEdge; i++)       // invert right edge
            {
                *((BYTE *)pData)++ = ~(*((BYTE *)pData));
            }
            memcpy(dstPtr, scanBuf1, dstWidthBytes);
            dstPtr += dstPitch;                 // next scan
        }
        break;

    case (0xaa):
        break;
    } // end: switch(rop3)
  } // endif: dst color key specified

  MODIFY_SLI_READ(ppdev, DISABLE_SLI_READ);

  pbd->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;

} // sli_DoBltNoSP

/*----------------------------------------------------------------------
Function name:  xsBltSrcToScanBuf

Description:    Apply color keys, then performs pixel replication in X,
                put result in stretch buffer.

Return:         void
----------------------------------------------------------------------*/
void __stdcall xsBltSrcToScanBuf(
    LPDDHAL_BLTDATA pbd, BYTE *srcPtr, BYTE *dstPtr,
    DWORD srcWidth, DWORD dstWidth,
    DWORD srcBytesPerPel, DWORD dstBytesPerPel)
{
    CKEY_TYPE colorKey;
    DWORD srcColorkeyMin, srcColorkeyMax;
    DWORD dstColorkeyMin, dstColorkeyMax, dwData;
    WORD  wData;
    int   twoSrcX, twoDstX, xErr;
    BYTE  *pSrc, *pDst, bData;
    DWORD srcWidthBytes, dstWidthBytes;

    colorKey = none;

    if (pbd->dwFlags & DDBLT_KEYSRCOVERRIDE)
    {
        colorKey = src;
        srcColorkeyMin = pbd->bltFX.ddckSrcColorkey.dwColorSpaceLowValue;
        srcColorkeyMax = pbd->bltFX.ddckSrcColorkey.dwColorSpaceHighValue;
    }

    if (pbd->dwFlags & DDBLT_KEYDESTOVERRIDE)
    {
        colorKey = (src==colorKey)? both : dst;
        dstColorkeyMin = pbd->bltFX.ddckDestColorkey.dwColorSpaceLowValue;
        dstColorkeyMax = pbd->bltFX.ddckDestColorkey.dwColorSpaceHighValue;
    }

    // replicate in X, result is in scan buffer 2

    srcWidthBytes = srcWidth * srcBytesPerPel;
    dstWidthBytes = dstWidth * dstBytesPerPel;
    twoSrcX = (int) (srcWidth + srcWidth);
    twoDstX = (int) (dstWidth + dstWidth);
    xErr = twoSrcX + (int)srcWidth - twoDstX;

    pSrc = (BYTE *)scanBuf1;
    pDst = (BYTE *)scanBuf2;
    memcpy(scanBuf1, srcPtr, srcWidthBytes); // read source into buffer 1
    memcpy(scanBuf2, dstPtr, dstWidthBytes); // read dst into buffer 2

    if (src == colorKey)    // do not write if src pixel matches color key
    {
        switch(srcBytesPerPel)
        {
        case 1: // 8bpp
            bData = *((BYTE *)pSrc);
            while (dstWidth--)
            {
                if ((bData < srcColorkeyMin) || (bData > srcColorkeyMax))
                {
                    *((BYTE *)pDst) = bData;
                }
                ((BYTE *)pDst)++;

                while (xErr >= 0)
                {
                    bData = * (++((BYTE *)pSrc));
                    xErr -= twoDstX;
                }
                xErr += twoSrcX;
            }
            break;

        case 2: // 16bpp
            wData = *((WORD *)pSrc);
            while (dstWidth--)
            {
                if ((wData < srcColorkeyMin) || (wData > srcColorkeyMax))
                {
                    *((WORD *)pDst) = wData;
                }
                ((WORD *)pDst)++;

                while (xErr >= 0)
                {
                    wData = * (++((WORD *)pSrc));
                    xErr -= twoDstX;
                }
                xErr += twoSrcX;
            }
            break;

        case 4: // 32bpp
            dwData = *((DWORD *)pSrc) & 0xffffff;
            while (dstWidth--)
            {
                if ((dwData < srcColorkeyMin) || (dwData > srcColorkeyMax))
                {
                    *((DWORD *)pDst) = dwData;
                }
                ((DWORD *)pDst)++;

                while (xErr >= 0)
                {
                    dwData = * (++((DWORD *)pSrc)) & 0xffffff;
                    xErr -= twoDstX;
                }
                xErr += twoSrcX;
            }
            break;

        } // end: switch(srcBytesPerPel)
    } // endif: use src color key
    else
    {
        if (dst == colorKey ) // update dst. pixel if matches color key
        {
            switch(srcBytesPerPel)
            {
            case 1: // 8bpp
                while (dstWidth--)
                {
                    bData = *((BYTE *)pDst);
                    if ((bData >= dstColorkeyMin) &&
                        (bData <= dstColorkeyMax)
                       )
                    {
                        *((BYTE *)pDst) = *((BYTE *)pSrc);
                    }
                    ((BYTE *)pDst)++;

                    while (xErr >= 0)
                    {
                        ((BYTE *)pSrc)++;
                        xErr -= twoDstX;
                    }
                    xErr += twoSrcX;
                }
                break;

            case 2: // 16bpp
                while (dstWidth--)
                {
                    wData = *((WORD *)pDst);
                    if ((wData >= dstColorkeyMin) &&
                        (wData <= dstColorkeyMax)
                       )
                    {
                        *((WORD *)pDst) = *((WORD *)pSrc);
                    }
                    ((WORD *)pDst)++;

                    while (xErr >= 0)
                    {
                        ((WORD *)pSrc)++;
                        xErr -= twoDstX;
                    }
                    xErr += twoSrcX;
                }
                break;

            case 4: // 32bpp
                while (dstWidth--)
                {
                    dwData = *((DWORD *)pDst);
                    if ((dwData >= dstColorkeyMin) &&
                        (dwData <= dstColorkeyMax)
                       )
                    {
                        *((DWORD *)pDst) = *((DWORD *)pSrc);
                    }
                    ((DWORD *)pDst)++;

                    while (xErr >= 0)
                    {
                        ((DWORD *)pSrc)++;
                        xErr -= twoDstX;
                    }
                    xErr += twoSrcX;
                }
                break;

            } // end: switch(bytesPerPel)
        } // endif: use dst color key
    } // endif: use src color key
}

/*----------------------------------------------------------------------
Function name:  sBltSrcToScanBuf

Description:    Performs pixel replication in X, put result in
                stretch buffer.

Return:         void
----------------------------------------------------------------------*/
void __stdcall sBltSrcToScanBuf(
    LPDDHAL_BLTDATA  pbd, BYTE *srcPtr, BYTE *dstPtr,
    DWORD srcWidth, DWORD dstWidth,
    DWORD srcBytesPerPel, DWORD dstBytesPerPel)
{
    int   twoSrcX, twoDstX, xErr;
    BYTE  *pSrc, *pDst;
    DWORD srcWidthBytes;

    // replicate in X, result is in scan buffer 2

    srcWidthBytes = srcWidth * srcBytesPerPel;
    twoSrcX = (int) (srcWidth + srcWidth);
    twoDstX = (int) (dstWidth + dstWidth);
    xErr = twoSrcX + (int)srcWidth - twoDstX;

    pSrc = (BYTE *)scanBuf1;
    pDst = (BYTE *)scanBuf2;
    memcpy(scanBuf1, srcPtr, srcWidthBytes);

    switch(srcBytesPerPel)
    {
    case 1: // 8bpp
        while (dstWidth--)
        {
            *((BYTE *)pDst)++ = *((BYTE *)pSrc);

            while (xErr >= 0)
            {
                ((BYTE *)pSrc)++;
                xErr -= twoDstX;
            }
            xErr += twoSrcX;
        }
        break;

    case 2: // 16bpp
        while (dstWidth--)
        {
            *((WORD *)pDst)++ = *((WORD *)pSrc);

            while (xErr >= 0)
            {
                ((WORD *)pSrc)++;
                xErr -= twoDstX;
            }
            xErr += twoSrcX;
        }
        break;

    case 4: // 32bpp
        while (dstWidth--)
        {
            *((DWORD *)pDst)++ = *((DWORD *)pSrc);

            while (xErr >= 0)
            {
                ((DWORD *)pSrc)++;
                xErr -= twoDstX;
            }
            xErr += twoSrcX;
        }
        break;

    } // end: switch(srcBytesPerPel)
}

/*----------------------------------------------------------------------
Function name:  sBltS

Description:    Performs pixel replication for stretch blt.

Return:         void
----------------------------------------------------------------------*/
void __stdcall sBltS(BLT_PARAMS)
{
    int   twoSrcY, twoDstY, yErr;
    DWORD srcWidth, dstWidth, srcHeight;
    PSBLTFUNC pfn;

    srcWidth = pbd->rSrc.right - pbd->rSrc.left;
    dstWidth = pbd->rDest.right - pbd->rDest.left;
    srcHeight = pbd->rSrc.bottom - pbd->rSrc.top;

    twoSrcY = (int) (srcHeight + srcHeight);
    twoDstY = (int) (dstHeight + dstHeight);
    yErr = twoSrcY + (int)srcHeight - twoDstY;

    if ( (pbd->dwFlags & DDBLT_KEYSRCOVERRIDE) ||
         (pbd->dwFlags & DDBLT_KEYDESTOVERRIDE) )
    {
        pfn = (PSBLTFUNC) &xsBltSrcToScanBuf;   // apply color keys
    }
    else
    {
        pfn = (PSBLTFUNC) &sBltSrcToScanBuf;    // no color keys
    }

    // stretch or shrink one scan, store result in scanline buffer

    (*pfn)(pbd, srcPtr, dstPtr, srcWidth, dstWidth, srcBytesPerPel, dstBytesPerPel);

    while (dstHeight--)
    {
        memcpy(dstPtr, scanBuf2, dstWidthBytes); // replicate in Y
        dstPtr += dstPitch;

        while (yErr >= 0)
        {
            srcPtr += srcPitch; // advance to next source scan
            yErr -= twoDstY;

            // stretch or shrink next scan, store result in scanline buffer

            (*pfn)(pbd, srcPtr, dstPtr, srcWidth, dstWidth, srcBytesPerPel, dstBytesPerPel);
        }
        yErr += twoSrcY;

    } // while (dstHeight--)
}

/*----------------------------------------------------------------------
Function name:  sli_DoBltS

Description:    Handle blts that use a source only, no pattern. Source
                can be from video memory or system memory. Destination
                is currently only in video memory.

Return:         DWORD DDRAW result

                DDHAL_DRIVER_HANDLED
----------------------------------------------------------------------*/

DWORD __stdcall
sli_DoBltS(NT9XDEVICEDATA   *ppdev,
           LPDDHAL_BLTDATA  pbd,
           DWORD            rop3,
           DWORD            srcPixelFormat,
           DWORD            dstPixelFormat,
           BLT_TYPE         blt_type)
{
  DWORD dstWidth, dstHeight, srcWidth, srcHeight;
  DWORD srcWidthBytes, dstWidthBytes;
  DWORD srcBytesPerPel, dstBytesPerPel, srcPitch, dstPitch, i;
  BYTE  *dstPtr, *srcPtr;
  LPDDRAWI_DDRAWSURFACE_LCL lpDstSurf, lpSrcSurf;

  FXBUSYWAIT(ppdev);

  MODIFY_SLI_READ(ppdev, ENABLE_SLI_READ);

  if (srcPixelFormat != dstPixelFormat)
  {
    // Software path cannot handle format conversion.

    pbd->ddRVal = DDERR_UNSUPPORTED;
    return DDHAL_DRIVER_HANDLED;
  }

  GET_BYTE_DEPTH(srcBytesPerPel, srcPixelFormat);
  GET_BYTE_DEPTH(dstBytesPerPel, dstPixelFormat);

  srcWidth = pbd->rSrc.right - pbd->rSrc.left;
  srcHeight = pbd->rSrc.bottom - pbd->rSrc.top;
  lpSrcSurf = pbd->lpDDSrcSurface;
  srcWidthBytes = srcWidth * srcBytesPerPel;

  if ( (vid2vid == blt_type) && IS_TILED(GET_HW_ADDR(lpSrcSurf)) )
  {
    srcPitch = _DS(ddTilePitch); // valid only if source is from video mem.
  }
  else
  {
    srcPitch = lpSrcSurf->lpGbl->lPitch;
  }

  dstWidth = pbd->rDest.right - pbd->rDest.left;
  dstHeight = pbd->rDest.bottom - pbd->rDest.top;
  lpDstSurf = pbd->lpDDDestSurface;
  dstWidthBytes = dstWidth * dstBytesPerPel;

  dstPitch = ( IS_TILED(GET_HW_ADDR(lpDstSurf)) ) ?
            _DS(ddTilePitch) : lpDstSurf->lpGbl->lPitch;

  // Check for overlaps. Since we are reading source into a scanline buffer
  // for performance reasons, we do not need to worry about overlaping in
  // the x direction.

  if ( (pbd->rDest.top >= pbd->rSrc.top) && (pbd->rDest.top <= pbd->rSrc.bottom) )
  {
    // copy from bottom up

    srcPtr = (BYTE *)
        ( lpSrcSurf->lpGbl->fpVidMem + ((pbd->rSrc.bottom - 1) * srcPitch) +
          (pbd->rSrc.left * srcBytesPerPel)
        );
#ifdef WINNT
    (DWORD)srcPtr += (DWORD)ppdev->pjLfbBase;
#endif
    dstPtr = (BYTE *)
        ( lpDstSurf->lpGbl->fpVidMem + ((pbd->rDest.bottom - 1) * dstPitch) +
          (pbd->rDest.left * dstBytesPerPel)
        );
#ifdef WINNT
    (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif
    srcPitch = 0 - srcPitch;
    dstPitch = 0 - dstPitch;
  }
  else
  {
    srcPtr = (BYTE *)
        ( lpSrcSurf->lpGbl->fpVidMem +
          (pbd->rSrc.top * srcPitch) + (pbd->rSrc.left * srcBytesPerPel)
        );
#ifdef WINNT
    (DWORD)srcPtr += (DWORD)ppdev->pjLfbBase;
#endif
    dstPtr = (BYTE *)
        ( lpDstSurf->lpGbl->fpVidMem +
          (pbd->rDest.top * dstPitch) + (pbd->rDest.left * dstBytesPerPel)
        );
#ifdef WINNT
    (DWORD)dstPtr += (DWORD)ppdev->pjLfbBase;
#endif
  }

  // check for stretch/shrink blt

  i = rop3 & 0xf;
  if ( (srcWidth == dstWidth) && (srcHeight == dstHeight) )
  {
    if ( (pbd->dwFlags & DDBLT_KEYSRCOVERRIDE) ||
         (pbd->dwFlags & DDBLT_KEYDESTOVERRIDE) )
    {
        (* xBltFuncTab[i])(pbd, dstWidthBytes, dstHeight, srcPitch, dstPitch,
            srcPtr, dstPtr, srcBytesPerPel, dstBytesPerPel);
    }
    else
    {
        (* BltFuncTab[i])(pbd, dstWidthBytes, dstHeight, srcPitch, dstPitch,
            srcPtr, dstPtr, srcBytesPerPel, dstBytesPerPel);
    }
  }
  else // stretch/shrink blt
  {
        (* sBltFuncTab[i])(pbd, dstWidthBytes, dstHeight, srcPitch, dstPitch,
            srcPtr, dstPtr, srcBytesPerPel, dstBytesPerPel);

  } // endif: non stretch/shrink blt

  MODIFY_SLI_READ(ppdev, DISABLE_SLI_READ);

  pbd->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;

} // sli_DoBltS

#endif